@@ -20,33 +20,45 @@ from typing import (
 
 import torch
 import torch.distributed as dist
+from pyre_extensions import none_throws
 
 from torcheval.metrics import Metric
-from torcheval.metrics.metric import TComputeReturn
-
-from torchtnt.utils import PGWrapper
-from typing_extensions import Literal
+from torcheval.metrics.metric import TComputeReturn, TState
+from torcheval.metrics.synclib import metrics_traversal_order, sync_states
 
 log: logging.Logger = logging.getLogger(__name__)
 
 _TMetrics = TypeVar("_TMetrics", bound=Iterable[Metric])
 
+_TMP: str = "tmp"
+
+
+def _get_world_size(process_group: Optional[dist.ProcessGroup]) -> int:
+    if not dist.is_available() or not dist.is_initialized():
+        # dist is not initialized or available, return 1 for world size
+        return 1
+    return dist.get_world_size(group=process_group)
+
+
+def _get_rank(process_group: Optional[dist.ProcessGroup]) -> int:
+    if not dist.is_available() or not dist.is_initialized():
+        # dist is not initialized or available, return 0 for rank
+        return 0
+    return dist.get_rank(group=process_group)
+
 
 def sync_and_compute(
     metric: Metric[TComputeReturn],
     process_group: Optional[dist.ProcessGroup] = None,
-    recipient_rank: Union[int, Literal["all"]] = 0,
-) -> Optional[TComputeReturn]:
+) -> TComputeReturn:
     """
     Sync metric states and returns the ``metric.compute()`` result of
-    synced metric on recipient rank. Return ``None`` on other ranks.
+    synced metric on all ranks.
 
     Args:
         metric: The metric object to be synced and computed.
         process_group: The process group on which the metric states are
             gathered. default: ``None`` (the entire world)
-        recipient_rank: The destination rank. If string "all" is passed in,
-            then all ranks are the destination ranks.
 
     Examples::
 
@@ -62,29 +74,11 @@ def sync_and_compute(
         tensor(2.) # Rank 2
         >>> sync_and_compute(max)
         tensor(2.) # Rank 0
-        None # Rank 1
-        None # Rank 2
-        >>> sync_and_compute(max, recipient_rank="all")
-        tensor(2.) # Rank 0
         tensor(2.) # Rank 1
         tensor(2.) # Rank 2
     """
-    # Sync metric states to rank 0, compute and broadcast the results
-    # when recipient_rank is "all".
-    # It uses less memory than syncing metric states to all ranks,
-    # since results are usually smaller in size than metric states.
-    dst_rank = 0 if recipient_rank == "all" else recipient_rank
-    synced_metric = get_synced_metric(metric, process_group, dst_rank)
-    compute_result = synced_metric.compute() if synced_metric else None
-
-    if recipient_rank == "all":
-        obj_list = [compute_result]
-        pg = PGWrapper(process_group)
-        if pg.get_rank() == dst_rank:
-            pg.broadcast_object_list(obj_list, src=int(dst_rank))
-        else:
-            pg.broadcast_object_list(obj_list, src=int(dst_rank))
-            compute_result = obj_list[0]
+    synced_metric = get_synced_metric(metric, process_group)
+    compute_result = synced_metric.compute()
 
     return compute_result
 
@@ -92,19 +86,15 @@ def sync_and_compute(
 def sync_and_compute_collection(
     metrics: MutableMapping[str, Metric],
     process_group: Optional[dist.ProcessGroup] = None,
-    recipient_rank: Union[int, Literal["all"]] = 0,
-) -> Optional[Dict[str, Any]]:
+) -> Dict[str, Any]:
     """
     Sync metric states across a dict of metrics and returns the
-    ``metric.compute()`` result of synced metrics on recipient rank.
-    Returns ``None`` on other ranks.
+    ``metric.compute()`` result of synced metrics on all ranks
 
     Args:
         metrics: The dict of metric objects to be synced and computed.
         process_group: The process group on which the metric states are
             gathered. default: ``None`` (the entire world)
-        recipient_rank: The destination rank. If string "all" is passed in,
-            then all ranks are the destination ranks.
 
     Examples::
 
@@ -124,32 +114,11 @@ def sync_and_compute_collection(
         tensor(2.) # Rank 2
         >>> sync_and_compute_collection(metrics)
         {"max" : tensor(2.), "min": tensor(0.)} # Rank 0
-        None # Rank 1
-        None # Rank 2
-        >>> sync_and_compute_collection(metrics, recipient_rank="all")
-        {"max" : tensor(2.), "min": tensor(0.)} # Rank 0
         {"max" : tensor(2.), "min": tensor(0.)} # Rank 1
         {"max" : tensor(2.), "min": tensor(0.)} # Rank 2
     """
-    # Sync metric states to rank 0, compute and broadcast the results
-    # when recipient_rank is "all".
-    # It uses less memory than syncing metric states to all ranks,
-    # since results are usually smaller in size than metric states.
-    dst_rank = 0 if recipient_rank == "all" else recipient_rank
-
-    synced_metrics = get_synced_metric_collection(metrics, process_group, dst_rank)
-    compute_result = None
-    if synced_metrics is not None:
-        compute_result = {key: m.compute() for key, m in synced_metrics.items()}
-
-    if recipient_rank == "all":
-        obj_list = [compute_result]
-        pg = PGWrapper(process_group)
-        if pg.get_rank() == dst_rank:
-            pg.broadcast_object_list(obj_list, src=int(dst_rank))
-        else:
-            pg.broadcast_object_list(obj_list, src=int(dst_rank))
-            compute_result = obj_list[0]
+    synced_metrics = get_synced_metric_collection(metrics, process_group)
+    compute_result = {key: m.compute() for key, m in synced_metrics.items()}
 
     return compute_result
 
@@ -157,18 +126,15 @@ def sync_and_compute_collection(
 def get_synced_state_dict(
     metric: Metric,
     process_group: Optional[dist.ProcessGroup] = None,
-    recipient_rank: Union[int, Literal["all"]] = 0,
 ) -> Dict[str, Any]:
     """
-    Return the state dict of a metric after syncing on recipient_rank.
+    Return the state dict of a metric after syncing on all ranks.
     Return an empty dict on other ranks.
 
     Args:
         metric: The metric object to sync and get ``state_dict()``
         process_group: The process group on which the metric states are
             gathered. default: ``None`` (the entire world)
-        recipient_rank: The destination rank. If string "all" is passed in,
-            then all ranks are the destination ranks.
     Returns:
         state dict of synced metric
 
@@ -183,32 +149,25 @@ def get_synced_state_dict(
         >>> max.update(torch.tensor(dist.get_rank()))
         >>> get_synced_state_dict(max)
         {"max", tensor(2.)} # Rank 0
-        {} # Rank 1
-        {} # Rank 2
-        >>> get_synced_state_dict(max, recipient_rank="all")
-        {"max", tensor(2.)} # Rank 0
         {"max", tensor(2.)} # Rank 1
         {"max", tensor(2.)} # Rank 2
     """
-    synced_metric = get_synced_metric(metric, process_group, recipient_rank)
+    synced_metric = get_synced_metric(metric, process_group)
     return synced_metric.state_dict() if synced_metric else {}
 
 
 def get_synced_state_dict_collection(
     metric_collection: MutableMapping[str, Metric],
     process_group: Optional[dist.ProcessGroup] = None,
-    recipient_rank: Union[int, Literal["all"]] = 0,
-) -> Optional[Dict[str, Dict[str, Any]]]:
+) -> Dict[str, Dict[str, Any]]:
     """
-    Return the state dict of a collection of metrics after syncing on recipient_rank.
+    Return the state dict of a collection of metrics after syncing on all ranks.
     Return an None on other ranks.
 
     Args:
         metric_collection (Dict[str, Metric]): The metric objects to sync and get ``state_dict()``
         process_group: The process group on which the metric states are
             gathered. default: ``None`` (the entire world)
-        recipient_rank: The destination rank. If string "all" is passed in,
-            then all ranks are the destination ranks.
     Returns:
         Bundle of state dicts of for the synced metrics
 
@@ -225,21 +184,14 @@ def get_synced_state_dict_collection(
         >>> minimum.update(torch.tensor(dist.get_rank()))
         >>> get_synced_state_dict({"max rank": maximum, "min rank": minimum})
         {"max rank": {"max", tensor(2.)}, "min rank": {"min", tensor(0.)}} # Rank 0
-        None # Rank 1
-        None # Rank 2
-        >>> get_synced_state_dict({"max rank": maximum, "min rank": minimum}, recipient_rank="all")
-        {"max rank": {"max", tensor(2.)}, "min rank": {"min", tensor(0.)}} # Rank 0
         {"max rank": {"max", tensor(2.)}, "min rank": {"min", tensor(0.)}} # Rank 1
         {"max rank": {"max", tensor(2.)}, "min rank": {"min", tensor(0.)}} # Rank 2
     """
     synced_metrics = get_synced_metric_collection(
         metric_collection,
         process_group,
-        recipient_rank,
     )
 
-    if synced_metrics is None:
-        return None
     return {key: metric.state_dict() for key, metric in synced_metrics.items()}
 
 
@@ -270,25 +222,15 @@ def clone_metrics(metrics: _TMetrics) -> List[Metric]:
 def get_synced_metric(
     metric: Metric,
     process_group: Optional[dist.ProcessGroup] = None,
-    recipient_rank: Union[int, Literal["all"]] = 0,
-) -> Optional[Metric]:
+) -> Metric:
     """
-    Returns a metric object on recipient_rank whose internal state
+    Returns a metric object on all ranks whose internal state
     variables are synced across processes in the process_group.
-    Returns ``None`` on non-recipient rank.
-
-    If ``all`` is passed as recipient_rank, all ranks in the
-    ``process_group`` are considered as recipient ranks.
 
     Args:
         metric: The metric object to sync.
         process_group: The process group on which the metric states are
             gathered. default: ``None`` (the entire world)
-        recipient_rank: The destination rank. If string "all" is passed in,
-            then all ranks are the destination ranks.
-    Raises:
-        ValueError: when ``recipient_rank`` is not an integer or string
-            "all".
 
     Examples::
 
@@ -307,67 +249,48 @@ def get_synced_metric(
         tensor(2.)     # Rank 0
         None # Rank 1 -- synced_metric is None
         None # Rank 2 -- synced_metric is None
-        >>> synced_metric = get_synced_metric(max, recipient_rank=1)
-        >>> synced_metric.compute() if synced_metric else None
-        None # Rank 0 -- synced_metric is None
-        tensor(2.)     # Rank 1
-        None # Rank 2 -- synced_metric is None
-        >>>  get_synced_metric(max, recipient_rank="all").compute()
+        >>> synced_metric = get_synced_metric(max)
+        >>> synced_metric.compute()
         tensor(2.) # Rank 0
         tensor(2.) # Rank 1
         tensor(2.) # Rank 2
     """
-    world_size = PGWrapper(process_group).get_world_size()
-    _validate_rank_and_world_size(recipient_rank, world_size)
+    world_size = _get_world_size(process_group)
+    _validate_rank_and_world_size(world_size)
 
     if world_size == 1:
         return metric
-    elif world_size == -1:
-        return None
 
     gathered_metric_list = _sync_metric_object(
         metric,
         # pyre-fixme[6]: For 2nd param expected `ProcessGroup` but got `Union[None,
         #  dist.ProcessGroup, _distributed_c10d.ProcessGroup]`.
         process_group if process_group else dist.group.WORLD,
-        recipient_rank,
         world_size,
     )
 
-    if gathered_metric_list is None:
-        return None
-    return (
-        clone_metric(gathered_metric_list[0])
-        .to(metric.device)
-        .merge_state(gathered_metric_list[1:])
-    )
+    local_rank = _get_rank(process_group)
+    other_rank_metrics: List[Metric] = [
+        gathered_metric_list[rank] for rank in range(world_size) if rank != local_rank
+    ]
+
+    return clone_metric(metric).to(metric.device).merge_state(other_rank_metrics)
 
 
 def get_synced_metric_collection(
     metric_collection: MutableMapping[str, Metric],
     process_group: Optional[dist.ProcessGroup] = None,
-    recipient_rank: Union[int, Literal["all"]] = 0,
-) -> Union[Optional[Dict[str, Metric]], MutableMapping[str, Metric]]:
+) -> Union[Dict[str, Metric], MutableMapping[str, Metric]]:
     """
-    Returns a dict of metric objects to the recipient_rank whose
+    Returns a dict of metric objects to all ranks whose
     internal state variables are synced across processes in the process_group.
-    Returns ``None`` on non-recipient rank.
-
 
     The data transfer is batched to maximize efficiency.
 
-    If ``all`` is passed as recipient_rank, all ranks in the
-    ``process_group`` are considered as recipient ranks.
-
     Args:
         metric_collection (Dict[str, Metric]): The dict of metric objects to sync.
         process_group (int): The process group on which the metric states are
             gathered. default: ``None`` (the entire world)
-        recipient_rank: The destination rank. If string "all" is passed in,
-            then all ranks are the destination ranks.
-    Raises:
-        ValueError: when ``recipient_rank`` is not an integer or string
-            "all".
 
     Examples::
 
@@ -380,19 +303,6 @@ def get_synced_metric_collection(
         >>> metrics["max"].update(torch.tensor(dist.get_rank()))
         >>> metrics["min"].update(torch.tensor(dist.get_rank()))
         >>> synced_metrics = get_synced_metric_collection(metrics)
-
-        by default metrics sync to Rank 0
-        >>> synced_metrics["max"].compute() if synced_metrics else None
-        tensor(2.) # Rank 0
-        None       # Rank 1 -- synced_metrics is None
-        None       # Rank 2 -- synced_metrics is None
-        >>> synced_metrics["min"].compute() if synced_metrics else None
-        tensor(0.) # Rank 0
-        None       # Rank 1 -- synced_metrics is None
-        None       # Rank 2 -- synced_metrics is None
-
-        you can also sync to all ranks or choose a specific rank
-        >>> synced_metrics = get_synced_metric_collection(metrics, recipient_rank="all")
         >>> synced_metrics["max"].compute()
         tensor(2.) # Rank 0
         tensor(2.) # Rank 1
@@ -402,71 +312,48 @@ def get_synced_metric_collection(
         tensor(0.) # Rank 1
         tensor(0.) # Rank 2
     """
-    world_size = PGWrapper(process_group).get_world_size()
-    _validate_rank_and_world_size(recipient_rank, world_size)
+    world_size = _get_world_size(process_group)
+    _validate_rank_and_world_size(world_size)
 
     if world_size == 1:
         return metric_collection
-    elif world_size == -1:
-        return None
 
     list_of_metric_collections = _sync_metric_object(
         metric_collection,
         # pyre-fixme[6]: For 2nd param expected `ProcessGroup` but got `Union[None,
         #  dist.ProcessGroup, _distributed_c10d.ProcessGroup]`.
         process_group if process_group else dist.group.WORLD,
-        recipient_rank,
         world_size,
     )
 
-    if list_of_metric_collections is None:
-        return None  # on non-recipient rank
+    if isinstance(list_of_metric_collections[0], MutableMapping):
+        local_rank = dist.get_rank(process_group)
 
-    elif isinstance(
-        list_of_metric_collections[0], MutableMapping
-    ):  # metric bundles are dicts.
+        # metric bundles are dicts.
         synced_metric_dict: Dict[str, Metric] = {}
-        # we use rank 0 metric to clone regardless of the recipient rank
-        rank_0_dict = list_of_metric_collections[0]
 
-        for metric_key in rank_0_dict.keys():
-            base_metric = rank_0_dict[metric_key]
+        for metric_key in metric_collection.keys():
+            base_metric = metric_collection[metric_key]
+            base_metric = clone_metric(base_metric).to(base_metric.device)
             other_rank_metrics: List[Metric] = [
                 list_of_metric_collections[rank][metric_key]
-                for rank in range(1, world_size)
+                for rank in range(world_size)
+                if rank != local_rank
             ]
-
-            synced_metric = (
-                clone_metric(base_metric)
-                .to(base_metric.device)
-                .merge_state(other_rank_metrics)
-            )
-
-            synced_metric_dict[metric_key] = synced_metric
-
+            synced_metric_dict[metric_key] = base_metric.merge_state(other_rank_metrics)
         return synced_metric_dict
 
 
 def _validate_rank_and_world_size(
-    recipient_rank: Union[int, Literal["all"]],
     world_size: int,
 ) -> None:
-    if not (isinstance(recipient_rank, int) or recipient_rank == "all"):
-        raise ValueError(
-            "``recipient_rank`` should be an integer or 'all', "
-            f"got {recipient_rank} instead."
-        )
-
     if world_size == 1:
         log.warning(
             "World size is 1, and metric(s) not synced. "
             "returning the input metric(s)."
         )
     elif world_size == -1:
-        log.warning(
-            "World size is -1, and current process might not be "
-            "in the process group. Returning ``None`` instead of synced metric(s)."
-        )
+        raise RuntimeError("The current process is not part of the process group")
     if world_size < 1:
         raise RuntimeError(
             f"Unexpected world_size {world_size} is seen when syncing metrics!"
@@ -475,60 +362,112 @@ def _validate_rank_and_world_size(
 
 @overload
 def _sync_metric_object(
-    metric_data: Metric,
+    local_metric_data: Metric,
     process_group: dist.ProcessGroup,
-    recipient_rank: Union[int, Literal["all"]],
     world_size: int,
-) -> Optional[List[Metric]]:
+) -> List[Metric]:
     ...
 
 
 @overload
 def _sync_metric_object(
-    metric_data: MutableMapping[str, Metric],
+    local_metric_data: MutableMapping[str, Metric],
     process_group: dist.ProcessGroup,
-    recipient_rank: Union[int, Literal["all"]],
     world_size: int,
-) -> Optional[List[Dict[str, Metric]]]:
+) -> List[MutableMapping[str, Metric]]:
     ...
 
 
+def _apply_device_to_tensor_states(
+    state_dict: Dict[str, Any], device: torch.device
+) -> None:
+    for state_name, state_value in state_dict.items():
+        if isinstance(state_value, torch.Tensor):
+            state_dict[state_name] = state_value.to(device)
+        elif isinstance(state_value, list):
+            state_dict[state_name] = [tensor.to(device) for tensor in state_value]
+        elif isinstance(state_value, dict):
+            state_dict[state_name] = {
+                key: tensor.to(device) for key, tensor in state_value.items()
+            }
+
+
 def _sync_metric_object(
-    metric_data: Union[Metric, MutableMapping[str, Metric]],
+    local_metric_data: Union[Metric, MutableMapping[str, Metric]],
     process_group: dist.ProcessGroup,
-    recipient_rank: Union[int, Literal["all"]],
     world_size: int,
-) -> Union[Optional[List[Metric]], Optional[List[Dict[str, Metric]]]]:
+) -> Union[List[Metric], List[MutableMapping[str, Metric]]]:
 
-    # Prepare metrics in data package for merge
-    if isinstance(metric_data, Metric):
-        metric_data._prepare_for_merge_state()
-    else:
-        for m in metric_data.values():
-            m._prepare_for_merge_state()
+    unpack = False  # unpack the dictionary into a single metric when returned. Only used when metric_data is a metric and not a dict of metrics.
+    if isinstance(local_metric_data, Metric):
+        local_metric_data = {_TMP: local_metric_data}
+        unpack = True
 
-    # create buffer for data to land in on recipient ranks
-    gathered_data_list = (
-        [None] * world_size
-        if recipient_rank == "all" or dist.get_rank() == recipient_rank
-        else None
+    # Allow metrics to run some pre-processing before syncing states
+    # for example, a common optimization is to concat tensors in a list into a single tensor
+    for m in local_metric_data.values():
+        m._prepare_for_merge_state()
+
+    # create a dict of state dicts for each metric in the collection, i.e. extract the state dicts from the Metric objects
+    metric_state_data: Dict[str, Dict[str, TState]] = {}
+    metric_to_device: Dict[str, torch.device] = {}
+    backend = dist.get_backend(group=process_group)
+    for metric_name, metric in local_metric_data.items():
+        metric_state_data[metric_name] = metric.state_dict()
+        if backend == "nccl" and metric.device.type == "cpu":
+            log.warning(
+                "Metric tensor states are on CPU, but NCCL process group detected. "
+                "These tensors will be moved to GPU prior to syncing. "
+                "It is recommended for efficiency reasons to either use a gloo process group to sync metrics on CPU, or move the metrics to GPU and use the NCCL process group."
+            )
+            # move tensors to gpu since on nccl
+            _apply_device_to_tensor_states(
+                metric_state_data[metric_name],
+                # pyre-ignore: Incompatible parameter type [6]
+                torch.cuda.current_device(),
+            )
+            # pyre-ignore: Incompatible parameter type [6]
+            metric_to_device[metric_name] = torch.cuda.current_device()
+        else:
+            metric_to_device[metric_name] = metric.device
+
+    metric_state_traversal_order = metrics_traversal_order(metric_state_data)
+
+    world_metric_data = sync_states(
+        metric_state_data,
+        metric_to_device,
+        metric_state_traversal_order,
+        process_group=process_group,
     )
+    world_metric_data = none_throws(world_metric_data)
 
-    # sync data
-    if isinstance(recipient_rank, int):
-        dist.gather_object(
-            metric_data,
-            gathered_data_list,
-            dst=recipient_rank,
-            group=process_group,
-        )
-    elif recipient_rank == "all":
-        dist.all_gather_object(gathered_data_list, metric_data, group=process_group)
-    # c10d does not have type annotations.
-    # pyre-ignore: Incompatible return type [7]: Expected `Union[List[Optional[Dict[str, Metric[typing.Any]]]], List[Optional[List[Metric[typing.Any]]]], List[Optional[Metric[typing.Any]]]]` but got `Optional[List[None]]`.
+    # Repack states into Metrics or Dict[str, Metric]s
+    if unpack:
+        # if users passed in one metric, read it from the "tmp" key and return a list of metrics.
+        gathered_data_list = []
+        for rank_data in world_metric_data:
+            gathered_data_list.append(_convert_to_psuedo_metric(rank_data[_TMP]))
+    else:
+        # if users passed in a dict[str, metric], return a list of dict[str, metric] populated with the gathered state dicts.
+        gathered_data_list = []
+        for rank_data in world_metric_data:
+            rank_dict = {}
+            for metric_name in local_metric_data.keys():
+                rank_dict[metric_name] = _convert_to_psuedo_metric(
+                    rank_data[metric_name]
+                )
+            gathered_data_list.append(rank_dict)
     return gathered_data_list
 
 
+# pyre-ignore: Missing return annotation [3]
+def _convert_to_psuedo_metric(metric_state_dict: Dict[str, Any]) -> Any:
+    """
+    Converts dictionary to object with attributes set according to key-value.
+    """
+    return type("", (), metric_state_dict)
+
+
 def reset_metrics(metrics: _TMetrics) -> _TMetrics:
     """
     Reset input metrics and returns the reset collection back to users.
